import sys

try:
    import urllib.parse as urlparse  # type: ignore
    from urllib.parse import urlencode  # type: ignore
except ImportError:  # python 2 fallback
    # pylint: disable=ungrouped-imports
    import urlparse  # type: ignore
    from urllib import urlencode  # type: ignore

if sys.version_info > (2, ):
    def encode_on_py3(arg, encoding):
        # type: (str, str) -> bytes
        return arg.encode(encoding)
else:
    def encode_on_py3(arg, encoding):
        # type: (str, str) -> bytes
        # pylint: disable=unused-argument
        return arg


def add_param_to_url(url, param_name, param_value):
    url_parts = list(urlparse.urlparse(url))
    query = dict(urlparse.parse_qsl(url_parts[4]))
    query[str(param_name)] = str(param_value)

    url_parts[4] = urlencode(query)
    return urlparse.urlunparse(url_parts)



import typing as t

from .utils import add_param_to_url


if t.TYPE_CHECKING:
    from .service_connector import ServiceConnector
    from mypy_extensions import TypedDict
    from typing_extensions import Literal

    _GroupsServiceData = TypedDict('_GroupsServiceData', {
        # Required data
        'context_groups_url': str,
        'scope': t.List[Literal['https://purl.imsglobal.org/spec/lti-gs/scope/contextgroup.readonly']],
        'service_versions': t.List[str],

        # Optional data
        'context_group_sets_url': str
    }, total=False)

    _Group = TypedDict('_Group', {
        # Required data
        'id': t.Union[str, int],
        'name': str,

        # Optional data
        'tag': str,
        'set_id': t.Union[str, int]
    }, total=False)

    _Set = TypedDict('_Set', {
        # Required data
        'id': t.Union[str, int],
        'name': str,

        # Optional data
        'groups': t.List[_Group]
    }, total=False)


class CourseGroupsService(object):
    _service_connector = None  # type: ServiceConnector
    _service_data = None  # type: _GroupsServiceData

    def __init__(self, service_connector, groups_service_data):
        # type: (ServiceConnector, _GroupsServiceData) -> None
        self._service_connector = service_connector
        self._service_data = groups_service_data

    def get_page(self, data_url, data_key='groups'):
        # type: (str, t.Optional[str]) -> t.Tuple[list, t.Optional[str]]
        """
        Get one page with the groups/sets.

        :param data_url
        :param data_key
        :return: tuple in format: (list with data items, next page url)
        """
        data = self._service_connector.make_service_request(
            self._service_data['scope'],
            data_url,
            accept='application/vnd.ims.lti-gs.v1.contextgroupcontainer+json',
        )
        data_body = t.cast(t.Any, data.get('body', {}))
        return data_body.get(data_key, []), data['next_page_url']

    def get_groups(self, user_id=None):
        groups_res_lst = []  # type: t.List[_Group]
        groups_url = self._service_data.get('context_groups_url')  # type: t.Optional[str]
        if user_id:
            groups_url = add_param_to_url(groups_url, 'user_id', user_id)

        while groups_url:
            groups, groups_url = self.get_page(groups_url, data_key='groups')
            groups_res_lst.extend(groups)

        return groups_res_lst

    def has_sets(self):
        return 'context_group_sets_url' in self._service_data

    def get_sets(self, include_groups=False):
        sets_res_lst = []  # type: t.List[_Set]
        sets_url = self._service_data.get('context_group_sets_url')  # type: t.Optional[str]

        while sets_url:
            sets, sets_url = self.get_page(sets_url, data_key='sets')
            sets_res_lst.extend(sets)

        if include_groups and sets_res_lst:
            set_id_to_index = {}
            for i, s in enumerate(sets_res_lst):
                set_id_to_index[s['id']] = i
                sets_res_lst[i]['groups'] = []

            groups = self.get_groups()
            for group in groups:
                set_id = group.get('set_id')
                if set_id and set_id in set_id_to_index:
                    index = set_id_to_index[set_id]
                    sets_res_lst[index]['groups'].append(group)

        return sets_res_lst



from ..exception import LtiException
from .abstract import MessageValidatorAbstract


class SubmissionReviewLaunchValidator(MessageValidatorAbstract):
    """Validates the body of a LTI submission review launch.

    The launch must include a for_user claim specifying the user
    who's submission is being reviewed, as well as the line item
    for the reviewed submission.
    """

    def validate(self, jwt_body):
        self.run_common_validators(jwt_body)

        if 'https://purl.imsglobal.org/spec/lti-ags/claim/endpoint' not in jwt_body:
            raise LtiException('Grade services must be included in a LtiSubmissionReviewRequest')

        ags_endpoint_claim = jwt_body['https://purl.imsglobal.org/spec/lti-ags/claim/endpoint']
        if 'lineitem' not in ags_endpoint_claim:
            raise LtiException('A LtiSubmissionReviewRequest must specify the lineitem it was launched for')

        for_user_claim = jwt_body.get('https://purl.imsglobal.org/spec/lti/claim/for_user')
        if for_user_claim is None:
            raise LtiException('For user claim must be included in a LtiSubmissionReviewRequest')
        if 'user_id' not in for_user_claim:
            raise LtiException('For user claim must include user_id')

        return True

    def can_validate(self, jwt_body):
        return jwt_body.get('https://purl.imsglobal.org/spec/lti/claim/message_type') == 'LtiSubmissionReviewRequest'
